/********************************COPYRIGHT*************************************

    Copyright (C) 2002 - 2005 Brendt Wohlberg  <software@wohlberg.net>

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 2 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this library; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

********************************COPYRIGHT*************************************/

/*
  This program extracts the scanner setting information embedded in
  TIFF files written by the NikonScan driver software for Nikon film
  scanners.

  Most recent modification: 5 April 2005
*/

#include <string.h>
#include <stdlib.h>
#if HAVE_BZLIB_H == 1 && HAVE_LIBBZ2 == 1
#include <bzlib.h>
#endif
#include "nsifd.h" 


void InitNSSettings(NSSettings* nss) {
  nss->date = 0;
  nss->software = 0;
  nss->hardware = 0;
  nss->mknver = 0;
  nss->filmtype = 0;
  nss->multsample = 0;
  nss->bitdepth = 0;
  nss->mastergain = 0;
  nss->colourgain = 0;
  nss->ice = 0;
  nss->roc = 0;
  nss->gem = 0;
}


void CleanNSSettings(NSSettings* nss) {
  free(nss->date);
  free(nss->software);
  free(nss->hardware);
  free(nss->mknver);
  free(nss->multsample);
  free(nss->bitdepth);
  free(nss->mastergain);
  free(nss->colourgain);
  free(nss->ice);
  free(nss->roc);
  free(nss->gem);
}


int NSSetDate(NSSettings* nss, const char* date) {
  static char mdate[128];
  char* cp;
  uint16 k;

  if ((nss->date = malloc(strlen(date)+1)) == 0)
    return 0;

  strncpy(mdate, date, 128);
  mdate[127] = '\0';
  cp = mdate;
  for (k=0; k<2; k++) {
    cp = strchr(cp, '.');
    *cp = '-';
  }
  for (k=0; k<2; k++) {
    cp = strchr(cp, '.');
    *cp = ':';
  }
  strcpy(nss->date, mdate);

  return 1;
}


int NSSetSoftware(NSSettings* nss, const char* software) {
  if ((nss->software = malloc(strlen(software)+1)) == 0)
    return 0;
  strcpy(nss->software, software);
  return 1;
}


int NSSetHardare(NSSettings* nss, const char* hardware) {
  char* cp;

  if ((nss->hardware = malloc(strlen(hardware)+1)) == 0)
    return 0;
  strcpy(nss->hardware, hardware);
  if ((cp = strstr(nss->hardware, "COOLSCAN")) != 0)
    strncpy(cp, "Coolscan", 8);
  return 1;
}


int NSSetMakerNoteVer(NSSettings* nss, const char* mknver) {
  if ((nss->mknver = malloc(strlen(mknver)+1)) == 0)
    return 0;
  strcpy(nss->mknver, mknver);
  return 1;
}


int NSSetFilmType(NSSettings* nss, const char* filmtype) {
  static const char* typelabels[] = {"Colour Negative", "B&W Negative", 
				     "Kodachrome", "Positive", "Negative"};

  if (strncmp(filmtype, "NEGATIVE(COLOR)", 15) == 0) {
    nss->filmtype = typelabels[0];
  } else if (strncmp(filmtype, "NEGATIVE(MONO)", 14) == 0) {
    nss->filmtype = typelabels[1];
  } else if (strncmp(filmtype, "KODACHROME", 10) == 0) {
    nss->filmtype = typelabels[2];
  } else if (strncmp(filmtype, "POSITIVE", 8) == 0) {
    nss->filmtype = typelabels[3];
  } else if (strncmp(filmtype, "NEGATIVE", 8) == 0) {
    nss->filmtype = typelabels[4];
  } else {
    return 0;
  }
  return 1;
}


int NSSetMultiSample(NSSettings* nss, const char* multsample) {
  if ((nss->multsample = malloc(strlen(multsample)+1)) == 0)
    return 0;
  strcpy(nss->multsample, multsample);
  return 1;
}


int NSSetBitDepth(NSSettings* nss, uint16 bitdepth) {
  if ((nss->bitdepth = malloc(sizeof(uint16))) == 0)
    return 0;
  *(nss->bitdepth) = bitdepth;
  return 1;
}


int NSSetMasterGain(NSSettings* nss, int32* mastergain, int fixflag) {
  if ((nss->mastergain = malloc(sizeof(float))) == 0)
    return 0;
  *(nss->mastergain) = (float)mastergain[0]/(float)mastergain[1];
  if (fixflag && *(nss->mastergain)<0.0)
    *(nss->mastergain) -= 0.01;
  return 1;
}


int NSSetColourGain(NSSettings* nss, int32* colourgain, int fixflag) {
  int16 k;

  if ((nss->colourgain = malloc(sizeof(float)*3)) == 0)
    return 0;
  for (k = 0; k < 3; k++) {
    nss->colourgain[k] = (float)colourgain[2*k]/(float)colourgain[2*k+1];
    if (fixflag && nss->colourgain[k]<0.0)
      nss->colourgain[k] -= 0.01;
  }
  return 1;
}


int NSSetICE(NSSettings* nss, const char* ice) {
  if ((nss->ice = malloc(strlen(ice)+1)) == 0)
    return 0;
  strcpy(nss->ice, ice);
  return 1;
}


int NSSetROC(NSSettings* nss, uint16 roc) {
  if ((nss->roc = malloc(sizeof(uint16))) == 0)
    return 0;
  *(nss->roc) = roc/10;
  return 1;
}


int NSSetGEM(NSSettings* nss, uint16 gem) {
  if ((nss->gem = malloc(sizeof(uint16))) == 0)
    return 0;
  *(nss->gem) = (gem<95)?((gem/20)-1):4;
  return 1;
}


void PrintNSSettings(FILE* fp, const NSSettings* nss) {
  if (nss->mknver)
    fprintf(fp, "MakerNote Version:  %s\n", nss->mknver);
  if (nss->date)
    fprintf(fp, "Scan Date:          %s\n", nss->date);
  if (nss->software)
    fprintf(fp, "Software:           %s\n", nss->software);
  if (nss->hardware)
    fprintf(fp, "Hardware:           %s\n", nss->hardware);
  if (nss->filmtype)
    fprintf(fp, "Film Type:          %s\n", nss->filmtype);
  if (nss->bitdepth)
    fprintf(fp, "Bit Depth:          %d\n", *(nss->bitdepth));
  if (nss->mastergain)
    fprintf(fp, "Analog Gain Master: %5.2f\n", *(nss->mastergain));
  if (nss->colourgain) {
    static const char* cstr[] = {"Red", "Green", "Blue"};
    static const char* sstr[] = {"  ", "", " "};
    int16 k;

    for (k = 0; k < 3; k++) {
      fprintf(fp, "Analog Gain %s:%s  %5.2f\n", cstr[k], sstr[k], 
	     nss->colourgain[k]);
    }
  }
  if (nss->ice)
    fprintf(fp, "Digital ICE:        %s\n", nss->ice);
  if (nss->roc)
    fprintf(fp, "Digital ROC:        %d\n", *(nss->roc));
  if (nss->gem)
    fprintf(fp, "Digital GEM:        %d\n", *(nss->gem));
  if (nss->multsample)
    fprintf(fp, "Multi Sample:       %s\n", nss->multsample);
}


#if HAVE_BZLIB_H == 1 && HAVE_LIBBZ2 == 1
FILE* BZUncompressed(const char* filename, uint32 bytes) {
  FILE* ifp;
  FILE* ofp;
  BZFILE* bp;
  int bzerror;
  const uint32 bufsize = 32768;
  uint32 nb, nr;
  char* buf = 0;
  
  if (!(ifp = fopen(filename, "r")))
    return 0;

  if ((buf = malloc(bufsize)) == 0) {
    fclose(ifp);
    return 0;
  }

  ofp = tmpfile();

  if (!(bp = BZ2_bzReadOpen(&bzerror, ifp, 0,0,0,0)) || (bzerror != BZ_OK)) {
    BZ2_bzReadClose(&bzerror, bp);
    fclose(ifp);
    free(buf);
    fclose(ofp);
    return 0;
  }

  bzerror = BZ_OK;
  while (bytes > 0 && bzerror == BZ_OK) {
    nb = (bytes > bufsize)?bufsize:bytes;
    nr = BZ2_bzRead(&bzerror, bp, buf, nb);
    if (bzerror == BZ_OK || bzerror == BZ_STREAM_END)
      fwrite(buf, 1, nr, ofp);
    else {
      BZ2_bzReadClose(&bzerror, bp);
      fclose(ifp);
      free(buf);
      fclose(ofp);
      return 0;
    }
    bytes -= nr;
  }

  BZ2_bzReadClose(&bzerror, bp);

  fclose(ifp);
  rewind(ofp);
  free(buf);

  return ofp;
}
#endif


int NSFileRead(const char* filename, NSSettings* nss, int fixflag) {
  FILE* fp = 0;
  TIFFHeader tiffhdr;
  uint8 swabflag = 0;
  TIFFDirEntry* tdep0 = 0;
  uint16 nde = 0;
  void *vp = 0;
  uint32 off, thoff;
  TIFFDirEntry* mkndep = 0;
  char mknver[5];
  TIFFDirEntry* htdep = 0;
  TIFFDirEntry* tdep = 0;
  uint16 k;
#if HAVE_BZLIB_H == 1 && HAVE_LIBBZ2 == 1
  char bz2magic[4];
#endif

  /* Open file */
  if (!(fp = fopen(filename, "r"))) {
    fprintf(stderr, "Error opening file %s\n", filename);
    return 1;
  }

#if HAVE_BZLIB_H == 1 && HAVE_LIBBZ2 == 1
    fread(bz2magic, 1, 3, fp);
    bz2magic[3] = '\0';
    if (strcmp(bz2magic, "BZh") == 0) {
      fclose(fp);
      /* Since the position of the private IFD appears to be
	 stable, uncompressing only the first 8192 bytes seems
	 to be an acceptable approach. */
      if (!(fp = BZUncompressed(filename, 8192))) {
	fprintf(stderr, "Error uncompressing bzip2 data\n");
	return 2;
      }
    }
    rewind(fp);
#endif

  /* Read TIFF header */
  if (fread(&tiffhdr, sizeof(tiffhdr), 1, fp) != 1) {
    fprintf(stderr, "Error reading TIFF header\n");
    return 3;
  }
  /* Check TIFF magic number */
  if (!(tiffhdr.tiff_magic == TIFF_BIGENDIAN ||
	tiffhdr.tiff_magic == TIFF_LITTLEENDIAN)) {
    fprintf(stderr, "Bad TIFF magic number: %x\n", tiffhdr.tiff_magic);
    return 4;
  }
  /* Determine whether endian correction necessary */
  swabflag = REQUIRE_ENDIAN_CORRECTION(tiffhdr.tiff_magic);

  /* Check additional TIFF magic number */
  if (swabflag)
    TIFFSwabShort(&tiffhdr.tiff_version);
  if (tiffhdr.tiff_version != TIFF_VERSION) {
    fprintf(stderr, "Bad TIFF version number: %x\n", tiffhdr.tiff_version);
    return 5;
  }

  /* Read directory entries for the first directory */
  tdep0 = OpenDirectory(fp, tiffhdr.tiff_diroff, &nde, swabflag);
  if (!tdep0) {
    fprintf(stderr, "Error reading first TIFF directory\n");
    return 6;
  }

  /* Get the value for the TIFFTAG_MAKE tag */
  vp =  TagValue(tdep0, nde, fp, TIFFTAG_MAKE, swabflag);
  if (!vp) {
    fprintf(stderr, "Error reading TIFFTAG_MAKE in first TIFF directory\n");
    return 7;
  }
  if (strcmp((char*)vp, "Nikon") != 0) {
    fprintf(stderr, "Not a Nikon generated TIFF file\n");
    return 8;
  }
  free(vp);

  /* Get the value for the ExifIFDPointer tag */
  vp =  TagValue(tdep0, nde, fp, 34665, swabflag);
  if (!vp) {
    fprintf(stderr, "Error reading ExifIFDPointer in first TIFF directory\n");
    return 9;
  }
  off = *(uint32*)vp;
  free(vp);

  /* Get the value for the DateTime tag */
  vp =  TagValue(tdep0, nde, fp, TIFFTAG_DATETIME, swabflag);
  if (vp && !NSSetDate(nss, (char*)vp))
    return 10;
  free(vp);

  /* Get the value for the Software tag */
  vp =  TagValue(tdep0, nde, fp, TIFFTAG_SOFTWARE, swabflag);
  if (vp && !NSSetSoftware(nss, (char*)vp))
    return 11;
  free(vp);

  /* Get the value for the Model tag */
  vp =  TagValue(tdep0, nde, fp, TIFFTAG_MODEL, swabflag);
  if (vp && !NSSetHardare(nss, (char*)vp)) 
    return 12;
  free(vp);

  /* Clean up*/
  CloseDirectory(tdep0);

  /* Read directory entries for the ExifIFDPointer directory */
  tdep0 = OpenDirectory(fp, off, &nde, swabflag);
  if (!tdep0) {
    fprintf(stderr, "Error reading ExifIFD TIFF directory\n");
    return 13;
  }

  /* Get the value for the MakerNote tag */
  tdep = TagDirEntry(tdep0, nde, 37500);
  if (!tdep) {
    fprintf(stderr, "Error reading MakerNote in ExifIFD TIFF directory\n");
    return 14;
  }
  off = tdep->tdir_offset;

  CloseDirectory(tdep0);

  /* Read directory entries for the MakerNote directory */
  mkndep = OpenMakerNote(fp, off, &nde, &thoff, swabflag);
  if (!mkndep) {
    fprintf(stderr, "Error reading MakerNote TIFF directory\n");
    return 15;
  }

  /* Get MakerNote version number */
  vp =  TagValue(mkndep, nde, fp, 0001, swabflag);
  memcpy(mknver, vp, 4);
  mknver[4] = '\0';
  free(vp);
  if (!NSSetMakerNoteVer(nss, mknver))
    return 16;

  /* MakerNote handling depends on version number */
  if (strcmp(mknver, "0100") == 0) {
    /* Version 0100 for NikonScan 3.12 (and probably earlier) */

    /* Get the value for tag 3600, providing the address of the
       private IFD containing the scanner settings. This address seems
       to always be 368. It is not clear whether the convoluted
       approach used here is really more reliable than just using this
       as a fixed offset for the position of private IFD. */
    vp =  TagValue(mkndep, nde, fp, 3600, swabflag);
    if (!vp) {
      fprintf(stderr, "Error reading tag 3600 in MakerNote TIFF directory\n");
      return 17;
    }
    off = *(uint32*)vp;
    free(vp);

    /* Read directory entries at position of private IFD */
    htdep = OpenDirectory(fp, off, &nde, swabflag);
    if (!htdep) {
      fprintf(stderr, "Error reading NikonScan private IFD\n");
      return 18;
    }
  } else if (strcmp(mknver, "0200") == 0) {
    /* Version 0200 for NikonScan 4.0 */

    /* Get a pointer to the IFD entry for tag 3593 */
    tdep = TagDirEntry(mkndep, nde, 3593);
    /* Get the value for tag 3593 */
    vp =  TagValue(mkndep, nde, fp, 3593, swabflag);
    /* Compute offset to private IFD containing scanner settings */
    off = tdep->tdir_offset + strlen((char*)vp) + 2;
    free(vp);

    /* Read directory entries at position of private IFD */
    htdep = OpenDirectory(fp, off, &nde, swabflag);
    if (!htdep) {
      fprintf(stderr, "Error reading NikonScan private IFD\n");
      return 19;
    }

    /* Work through each directory entry modifying the offsets to be
       relative to the start of the file rather than the start of the
       MakerNote TIFF header */
    for (k = 0; k < nde; k++) {
      if (DirEntrySize(&htdep[k]) > 4)
	htdep[k].tdir_offset += thoff;
    }
  } else {
    fprintf(stderr, "Unrecognised MakerNote version number\n");
    return 20;
  }

  CloseDirectory(mkndep);


 /* Work through each directory entry */
  for (k = 0; k < nde; k++) {
    /* Get directory entry value and assign appropriate meaning */
    if ((vp = DirEntryValue(&htdep[k],fp,swabflag)) == 0) {
      fprintf(stderr, "Error getting value of directory entry 0x%x\n",
	      htdep[k].tdir_tag);
      return 21;
    }

    switch (htdep[k].tdir_tag) {
    case   2: if (!NSSetFilmType(nss, (char*)vp)) return 21;
      break;
    case  64: if (!NSSetMultiSample(nss, (char*)vp)) return 22;
      break;
    case  65: if (!NSSetBitDepth(nss, *(uint16*)vp)) return 23;
      break;
    case  80: if (!NSSetMasterGain(nss, (int32*)vp, fixflag)) return 24;
      break;
    case  81: if (!NSSetColourGain(nss, (int32*)vp, fixflag)) return 25;
      break;
    case 256: if (!NSSetICE(nss, (char*)vp)) return 26;
      break;
    case 272: if (!NSSetROC(nss, (uint16)*(uint32*)vp)) return 27;
      break;
    case 288: if (!NSSetGEM(nss, (uint16)*(uint32*)vp)) return 28;
      break;
    default: fprintf(stderr, "Warning: encountered unrecognised tag 0x%x\n",
		     htdep[k].tdir_tag);
    }
    free(vp);
  }

  /* Clean up */
  if (htdep)
    CloseDirectory(htdep);
  fclose(fp);
  
  return 0;
}
